#!/usr/bin/env python3

import unittest

from src.pg_lock_tracer.helper import PostgreSQLLockHelper


class UNITTests(unittest.TestCase):
    def test_encode_locks(self):
        """
        Test the encoding of locks
        """
        my_locks = [
            "NoLock",
            "AccessShareLock",
            "ShareRowExclusiveLock",
            "AccessExclusiveLock",
        ]

        my_locks_numeric = list(map(PostgreSQLLockHelper.lock_type_to_int, my_locks))

        # Test conversion was ok
        self.assertListEqual([0, 1, 6, 8], my_locks_numeric)

    def encode_and_decode_locks(self, locks):
        """
        Decode and encode the list of locks and test for errors
        """

        # Convert into numeric values
        my_locks_numeric = list(map(PostgreSQLLockHelper.lock_type_to_int, locks))

        # Encode and decode into single value
        single_value = PostgreSQLLockHelper.encode_locks_into_value(my_locks_numeric)
        decoded_locks = PostgreSQLLockHelper.decode_locks_from_value(single_value)

        decoded_my_locks = list(
            map(PostgreSQLLockHelper.lock_type_to_str, decoded_locks)
        )
        self.assertListEqual(locks, decoded_my_locks)

    def test_parse_lock_decoding_and_encoding0(self):
        """
        Test encoding and decoding of locks into a single value
        """

        my_locks = [
            "AccessExclusiveLock",
        ]

        self.encode_and_decode_locks(my_locks)

    def test_parse_lock_decoding_and_encoding1(self):
        """
        Test encoding and decoding of locks into a single value
        """

        my_locks = [
            "NoLock",
            "AccessShareLock",
            "ShareRowExclusiveLock",
            "AccessExclusiveLock",
        ]

        self.encode_and_decode_locks(my_locks)

    def test_parse_lock_decoding_and_encoding2(self):
        """
        Test encoding and decoding of locks into a single value
        """

        my_locks = ["NoLock"]

        self.encode_and_decode_locks(my_locks)

    def test_parse_lock_decoding_and_encoding3(self):
        """
        Test encoding and decoding of locks into a single value
        """

        my_locks = []

        self.encode_and_decode_locks(my_locks)

    def test_parse_lock_decoding_and_encoding_dup(self):
        """
        Test encoding and decoding with duplicates
        """

        my_locks = [
            "AccessShareLock",
            "AccessShareLock",
            "AccessShareLock",
            "ShareRowExclusiveLock",
            "ShareRowExclusiveLock",
        ]

        my_locks_numeric = list(map(PostgreSQLLockHelper.lock_type_to_int, my_locks))

        # Encode and decode into single value
        single_value = PostgreSQLLockHelper.encode_locks_into_value(my_locks_numeric)
        decoded_locks = PostgreSQLLockHelper.decode_locks_from_value(single_value)

        decoded_my_locks = list(
            map(PostgreSQLLockHelper.lock_type_to_str, decoded_locks)
        )

        # Only unique values are present
        self.assertListEqual([my_locks[0], my_locks[4]], decoded_my_locks)



"""Resolve PostgreSQL OIDs to names and cache the result"""

import sys

from urllib.parse import urlparse

import psycopg2


class OIDResolver:
    def __init__(self, connection_url):
        self.connection_url = connection_url
        self.cache = {}
        self.connection = None
        self.cur = None
        self.connect()

    def connect(self):
        """
        Open the database connection
        """
        connection_url_parsed = urlparse(self.connection_url)
        username = connection_url_parsed.username
        password = connection_url_parsed.password
        database = connection_url_parsed.path[1:]
        hostname = connection_url_parsed.hostname
        port = connection_url_parsed.port

        try:
            self.connection = psycopg2.connect(
                database=database,
                user=username,
                password=password,
                host=hostname,
                port=port,
            )
            self.connection.set_session(autocommit=True)
            self.cur = self.connection.cursor()

            # Warmup cache
            self.fetch_all_oids()
        except psycopg2.OperationalError as error:
            print(f"Unable to connect to the database {self.connection_url}")
            print(f"{error}")
            sys.exit(1)

    def disconnect(self):
        """
        Close the database connection.
        """
        if self.cur:
            self.cur.close()
            self.cur = None

        if self.connection:
            self.connection.close()
            self.connection = None

    def fetch_all_oids(self):
        """
        Fetch all Oid mappings from the catalog and cache them. This
        is done because:

        (1) Cache Oid cache lookups have to be fast and we want
            a warm cache.

        (2) Operations such as DROP delete objects from the database.
            Fetching the oid mapping afterwards is not possible.
        """

        select_stmt = """
        SELECT n.nspname, c.relname, c.oid 
        FROM pg_namespace n
        JOIN pg_class c ON n.oid = c.relnamespace
        """

        self.cur.execute(select_stmt)

        oids = self.cur.fetchall()

        for result_row in oids:
            oid = result_row[2]
            name = f"{result_row[0]}.{result_row[1]}"
            self.cache[oid] = name

    def fetch_oid_from_db(self, oid):
        """
        Resolve the given OID into a name
        """

        select_stmt = """
        SELECT n.nspname, c.relname
        FROM pg_namespace n
        JOIN pg_class c ON n.oid = c.relnamespace
        WHERE c.oid = %s;
        """

        oid = str(oid)

        try:
            self.cur.execute(
                select_stmt,
                [
                    oid,
                ],
            )

            result_row = self.cur.fetchone()

            # Unable to get name, return Oid instead
            if result_row is None:
                return f"Oid {oid}"

            name = f"{result_row[0]}.{result_row[1]}"

            # Cache result
            self.cache[oid] = name

            return name
        except psycopg2.Error as error:
            print(f"Error while executing SQL statement: {error}")
            print(f"pgerror: {error.pgerror}")
            print(f"pgcode: {error.pgcode}")
            return ""

    def resolve_oid(self, oid):
        """
        Resolve the given OID into a name.
        """

        # OID cache hit
        if oid in self.cache:
            return self.cache[oid]

        # OID cache miss
        return self.fetch_oid_from_db(oid)



#!/usr/bin/env python3
#
# PostgreSQL row lock tracer.
#
# See https://www.postgresql.org/docs/current/explicit-locking.html#LOCKING-ROWS
#
###############################################

import sys
import argparse

from enum import IntEnum, unique

from bcc import BPF
from prettytable import PrettyTable

from pg_lock_tracer import __version__
from pg_lock_tracer.helper import BPFHelper

EXAMPLES = """examples:
# Trace the row locks of the given PostgreSQL binary
pg_row_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace the row locks of the PID 1234
pg_row_lock_tracer -p 1234 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace the row locks of the PID 1234 and 5678
pg_row_lock_tracer -p 1234 -p 5678 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace the row locks of the PID 1234 and be verbose
pg_row_lock_tracer -p 1234 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres -v

# Trace the row locks and show statistics
pg_row_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres --statistics
"""

parser = argparse.ArgumentParser(
    description="",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=EXAMPLES,
)
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"{parser.prog} ({__version__})",
)
parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
parser.add_argument(
    "-p",
    "--pid",
    type=int,
    nargs="+",
    dest="pids",
    metavar="PID",
    help="the pid(s) to trace",
)
parser.add_argument(
    "-x",
    "--exe",
    type=str,
    required=True,
    dest="path",
    metavar="PATH",
    help="path to binary",
)
parser.add_argument(
    "-d",
    "--dry-run",
    action="store_true",
    help="compile and load the BPF program but exit afterward",
)
parser.add_argument("--statistics", action="store_true", help="print lock statistics")


@unique
class Events(IntEnum):
    LOCK_TUPLE = 0
    LOCK_TUPLE_END = 1


# See lockoptions.h in PostgreSQL
@unique
class TMResult(IntEnum):
    TM_OK = 0
    TM_INVISIBLE = 1
    TM_SELFMODIFIED = 2
    TM_UPDATED = 3
    TM_DELETED = 4
    TM_BEINGMODIFIED = 5
    TM_WOULDBLOCK = 6


# See lockoptions.h in PostgreSQL
@unique
class LockWaitPolicy(IntEnum):
    LOCK_WAIT_BLOCK = 0
    LOCK_WAIT_SKIP = 1
    LOCK_WAIT_ERROR = 2


# See lockoptions.h in PostgreSQL
@unique
class LockTupleMode(IntEnum):
    LOCK_TUPLE_KEYSHARE = 0
    LOCK_TUPLE_SHARE = 1
    LOCK_TUPLE_NOKEYEXCLUSIVE = 2
    LOCK_TUPLE_EXCLUSIVE = 3


class LockStatisticsEntry:
    def __init__(self) -> None:
        # The requested locks
        self._lock_modes = {}

        # The used lock policies
        self._lock_policies = {}

        # The lock results
        self._lock_results = {}

    @property
    def lock_modes(self):
        return self._lock_modes

    @lock_modes.setter
    def lock_modes(self, value):
        self._lock_modes = value

    @property
    def lock_policies(self):
        return self._lock_policies

    @lock_policies.setter
    def lock_policies(self, value):
        self._lock_policies = value

    @property
    def lock_results(self):
        return self._lock_results

    @lock_results.setter
    def lock_results(self, value):
        self._results = value


class PGRowLockTracer:
    def __init__(self, prog_args):
        self.bpf_instance = None
        self.args = prog_args
        self.statistics = {}

        # Variables for lock timing
        self.last_lock_request_time = {}

        # Belong the processes to the binary?
        BPFHelper.check_pid_exe(self.args.pids, self.args.path)

    def get_lock_wait_time(self, event):
        """
        Get the last lock wait time (WAIT_START updates
        last_lock_request_time).
        """
        if event.event_type != Events.LOCK_TUPLE_END:
            return None

        return event.timestamp - self.last_lock_request_time[event.pid]

    def update_statistics(self, event):
        """
        Update the statistics
        """
        if event.pid not in self.statistics:
            self.statistics[event.pid] = LockStatisticsEntry()

        statistics_entry = self.statistics.get(event.pid)

        # Lock requested
        if event.event_type == Events.LOCK_TUPLE:
            lock_wait_policy = LockWaitPolicy(event.lockwaitpolicy)

            if lock_wait_policy in statistics_entry.lock_policies:
                statistics_entry.lock_policies[lock_wait_policy] += 1
            else:
                statistics_entry.lock_policies[lock_wait_policy] = 1

            lock_tuple_mode = LockTupleMode(event.locktuplemode)

            if lock_tuple_mode in statistics_entry.lock_modes:
                statistics_entry.lock_modes[lock_tuple_mode] += 1
            else:
                statistics_entry.lock_modes[lock_tuple_mode] = 1

            return

        # Lock request done
        if event.event_type == Events.LOCK_TUPLE_END:
            lock_result = TMResult(event.lockresult)

            if lock_result in statistics_entry.lock_results:
                statistics_entry.lock_results[lock_result] += 1
            else:
                statistics_entry.lock_results[lock_result] = 1
            return

        return

    def print_lock_event(self, _cpu, data, _size):
        """
        Print a new lock event.
        """
        event = self.bpf_instance["lockevents"].event(data)

        if self.args.pids and event.pid not in self.args.pids:
            return

        print_prefix = f"{event.timestamp} [Pid {event.pid}]"

        self.update_statistics(event)

        if event.event_type == Events.LOCK_TUPLE:
            self.last_lock_request_time[event.pid] = event.timestamp

            locktuplemode = LockTupleMode(event.locktuplemode).name
            lockwaitpolicy = LockWaitPolicy(event.lockwaitpolicy).name

            print(
                f"{print_prefix} LOCK_TUPLE (Tablespace {event.tablespace} "
                f"database {event.database} relation {event.relation}) "
                f"- (Block and offset {event.blockid} {event.offset}) "
                f"- {locktuplemode} {lockwaitpolicy}"
            )
        elif event.event_type == Events.LOCK_TUPLE_END:
            lockresult = TMResult(event.lockresult).name
            needed_time = self.get_lock_wait_time(event)
            print(f"{print_prefix} LOCK_TUPLE_END {lockresult} in {needed_time} ns")
        else:
            raise ValueError(f"Unknown event type {event.event_type}")

    def init(self):
        """
        Init the PostgreSQL lock tracer
        """
        enum_defines = BPFHelper.enum_to_defines(Events, "EVENT")

        bpf_program = BPFHelper.read_bpf_program("pg_row_lock_tracer.c")
        bpf_program_final = bpf_program.replace("__DEFINES__", enum_defines)

        if self.args.verbose:
            print(bpf_program_final)

        # Disable warnings like
        # 'warning: '__HAVE_BUILTIN_BSWAP32__' macro redefined [-Wmacro-redefined]'
        bpf_cflags = ["-Wno-macro-redefined"] if not self.args.verbose else []

        print("===> Compiling BPF program")
        self.bpf_instance = BPF(text=bpf_program_final, cflags=bpf_cflags)

        print("===> Attaching BPF probes")
        self.attach_probes()

        # Open the event queue
        self.bpf_instance["lockevents"].open_perf_buffer(
            self.print_lock_event, page_cnt=BPFHelper.page_cnt
        )

    def attach_probes(self):
        """
        Attach BPF probes
        """
        BPFHelper.register_ebpf_probe(
            self.args.path,
            self.bpf_instance,
            "^heapam_tuple_lock$",
            "heapam_tuple_lock",
            self.args.verbose,
        )
        BPFHelper.register_ebpf_probe(
            self.args.path,
            self.bpf_instance,
            "^heapam_tuple_lock$",
            "heapam_tuple_lock_end",
            self.args.verbose,
            False,
        )

    def print_statistics(self):
        """
        Print lock statistics
        """
        print("\nLock statistics:\n================")

        # Wait policies
        print("\nUsed wait policies:")
        wait_polices = ["PID"]

        for wait_policy in LockWaitPolicy:
            wait_polices.append(wait_policy.name)

        table = PrettyTable(wait_polices)

        for pid in sorted(self.statistics):
            statistics = self.statistics[pid]
            pid_statistics = [pid]
            for wait_policy in LockWaitPolicy:
                pid_statistics.append(statistics.lock_policies.get(wait_policy, 0))
            table.add_row(pid_statistics)
        print(table)

        # Lock modes
        print("\nLock modes:")
        lock_modes = ["PID"]

        for lock_mode in LockTupleMode:
            lock_modes.append(lock_mode.name)

        table = PrettyTable(lock_modes)

        for pid in sorted(self.statistics):
            statistics = self.statistics[pid]
            pid_statistics = [pid]
            for lock_mode in LockTupleMode:
                pid_statistics.append(statistics.lock_modes.get(lock_mode, 0))
            table.add_row(pid_statistics)
        print(table)

        # Lock results
        print("\nLock results:")
        lock_results = ["PID"]

        for lock_result in TMResult:
            lock_results.append(lock_result.name)

        table = PrettyTable(lock_results)

        for pid in sorted(self.statistics):
            statistics = self.statistics[pid]
            pid_statistics = [pid]
            for lock_result in TMResult:
                pid_statistics.append(statistics.lock_results.get(lock_result, 0))
            table.add_row(pid_statistics)
        print(table)

    def run(self):
        """
        Run the BPF program and read results
        """
        print("===> Ready to trace")
        while True:
            try:
                self.bpf_instance.perf_buffer_poll()
            except KeyboardInterrupt:
                if self.args.statistics:
                    self.print_statistics()
                sys.exit(0)


def main():
    """
    Entry point for the BPF based PostgreSQL row lock tracer.
    """
    args = parser.parse_args()

    pg_lock_tracer = PGRowLockTracer(args)
    pg_lock_tracer.init()

    if not args.dry_run:
        pg_lock_tracer.run()


if __name__ == "__main__":
    main()



"""
Helper classes
"""

import os

from pathlib import Path

from bcc import BPF


class PostgreSQLLockHelper:
    """
    # Defines taken from: src/include/storage/lockdefs.h
    #
    # NoLock is not a lock mode, but a flag value meaning "don't get a lock"
    # define NoLock                  0
    #
    # define AccessShareLock         1   /* SELECT */
    # define RowShareLock            2   /* SELECT FOR UPDATE/FOR SHARE */
    # define RowExclusiveLock        3   /* INSERT, UPDATE, DELETE */
    # define ShareUpdateExclusiveLock 4  /* VACUUM (non-FULL), ANALYZE, CREATE
    #                                    * INDEX CONCURRENTLY */
    # define ShareLock               5   /* CREATE INDEX (WITHOUT CONCURRENTLY) */
    # define ShareRowExclusiveLock   6   /* like EXCLUSIVE MODE, but allows ROW
    #                                    * SHARE */
    # define ExclusiveLock           7   /* blocks ROW SHARE/SELECT...FOR UPDATE */
    # define AccessExclusiveLock     8   /* ALTER TABLE, DROP TABLE, VACUUM FULL,
    #                                    * and unqualified LOCK TABLE */
    """

    locks = {}
    locks["NoLock"] = 0
    locks["AccessShareLock"] = 1
    locks["RowShareLock"] = 2
    locks["RowExclusiveLock"] = 3
    locks["ShareUpdateExclusiveLock"] = 4
    locks["ShareLock"] = 5
    locks["ShareRowExclusiveLock"] = 6
    locks["ExclusiveLock"] = 7
    locks["AccessExclusiveLock"] = 8

    @staticmethod
    def encode_locks_into_value(locks):
        """
        Encode a list of lock types into a single value
        """
        result = 0

        for lock in locks:
            lock_value = 1 << int(lock)
            result |= lock_value

        return result

    @staticmethod
    def decode_locks_from_value(encoded_value):
        """
        Decode a value of locks into a list of locks
        """
        result = []

        for lock in range(0, 9):
            lock_value = 1 << int(lock)

            if encoded_value & lock_value == lock_value:
                result.append(lock)

        return result

    @staticmethod
    def lock_type_to_str(lock_type):
        """
        Return the name of a lock based on the numeric value
        """
        for lock_name, lock_numeric_value in PostgreSQLLockHelper.locks.items():
            if lock_numeric_value == lock_type:
                return lock_name

        raise ValueError(f"Unsupported lock type {lock_type}")

    @staticmethod
    def lock_type_to_int(lock_name):
        """
        Return the numeric value of a lock based on the name
        """

        if lock_name not in PostgreSQLLockHelper.locks:
            raise ValueError(f"Unknown lock type {lock_name}")

        return PostgreSQLLockHelper.locks[lock_name]


class BPFHelper:
    # The size of the kernel ring buffer
    page_cnt = 2048

    @staticmethod
    def enum_to_defines(enum_instance, prefix):
        """
        Convert a IntEnum into C '#define' statements
        """
        result = ""

        for instance in enum_instance:
            result += f"#define {prefix}_{instance.name} {instance.value}\n"

        return result

    @staticmethod
    def read_bpf_program(program_name):
        """
        Load the BPF program from a file and return it as a string.
        """
        program_file = Path(__file__).parent / "bpf" / program_name

        if not os.path.exists(program_file):
            raise ValueError(f"BPF program file not found {program_file}")

        with program_file.open("r") as bpf_program:
            return bpf_program.read()

    @staticmethod
    def check_pid_exe(pids, executable):
        """
        Do the given PIDs belong to the executable
        """
        if not pids:
            return

        for pid in pids:
            if not os.path.isdir(f"/proc/{pid}"):
                raise ValueError(
                    f"/proc entry for pid {pid} not found, does the process exist?"
                )

            binary = os.readlink(f"/proc/{pid}/exe")

            if binary != executable:
                raise ValueError(
                    f"Pid {pid} does not belong to binary {executable}. Executable is {binary}"
                )

    @staticmethod
    def register_ebpf_probe(
        path, bpf_instance, function_regex, bpf_fn_name, verbose, probe_on_enter=True
    ):
        """
        Register a BPF probe
        """
        addresses = set()
        func_and_addr = BPF.get_user_functions_and_addresses(path, function_regex)

        if not func_and_addr:
            raise ValueError(f"Unable to locate function {function_regex}")

        # Handle address duplicates
        for function, address in func_and_addr:
            if address in addresses:
                continue
            addresses.add(address)

            if probe_on_enter:
                bpf_instance.attach_uprobe(name=path, sym=function, fn_name=bpf_fn_name)
                if verbose:
                    print(f"Attaching to {function} at address {address} on enter")
            else:
                bpf_instance.attach_uretprobe(
                    name=path, sym=function, fn_name=bpf_fn_name
                )
                if verbose:
                    print(f"Attaching to {function} at address {address} on return")



#!/usr/bin/env python3
#
# Generate an animated version of the locks
# of a query. Input data is the JSON output of
# pg_lock_tracker.
#
# Current limitations:
# * Only lock traces of one query are supported.
##########################

import os
import json
import argparse

import igraph
import graphviz

from pg_lock_tracer import __version__
from pg_lock_tracer.helper import PostgreSQLLockHelper

# See https://github.com/magjac/d3-graphviz/blob/master/examples/basic-unpkg-worker.html
HTML_TEMPLATE = """
<!DOCTYPE html>
<meta charset="utf-8">
<body>
<script src="https://d3js.org/d3.v7.js"></script>
<script src="https://unpkg.com/@hpcc-js/wasm@2/dist/graphviz.umd.js"></script>
<script src="https://unpkg.com/d3-graphviz@5/build/d3-graphviz.js"></script>

<style>
.frame_div {
   display: inline-block;
}

#frame_control {
   background-color: lightgray;
   height: 24px;
   margin-bottom: 5px;
   padding-left: 5px;
}

#pause_control {
   padding-left: 20px;
}

input, output {
  vertical-align: middle;
}
</style>
<div id="frame_control">
  Frame <input type="range" id="active_frame_slider" min="0" max="100" value="0" oninput="updateSlider(this.value);" onchange="render();">
  <div id="frames" class="frame_div">
     (<div id="cur_frame" class="frame_div"></div> of <div id="total_frame" class="frame_div"></div>)
  </div>
  <div id="pause_control" class="frame_div">
    <input type="checkbox" id="render_active" name="render_active" value="active" checked onclick="render();">
     <label for="render_active" style="vertical-align: bottom;">&#9658; Play</label>
    </input>
  </div>
</div>

<div id="graph" style="text-align: center;"></div>

<script>
var dotIndex = 0;

var graphviz = d3.select("#graph").graphviz()
    .attributer(attributer)
    .transition(function () {
        return d3.transition("main")
            .ease(d3.easeLinear)
            .delay(500)
            .duration(1500);
    })
    .logEvents(true)
    .engine('circo')
    .on("initEnd", render);

function render() {
    var dotLines = dots[dotIndex];
    var dot = dotLines.join('');
    graphviz
        .renderDot(dot)
        .on("end", function () {
            if (! document.getElementById('render_active').checked) {
                return;
            }
            dotIndex = (dotIndex + 1) % dots.length;
            updateFrames();
            document.getElementById("active_frame_slider").value = dotIndex;
            render();
        })
        .zoom(true);
}

function updateFrames() {
    document.getElementById("cur_frame").innerHTML = dotIndex;
    document.getElementById("total_frame").innerHTML = dots.length;
}

function updateSlider(newValue) {
    dotIndex = Number(newValue);
    updateFrames();
}

function initSlider() {
    document.getElementById("active_frame_slider").value = dotIndex;
    document.getElementById("active_frame_slider").max = dots.length;
}

function attributer(datum, index, nodes) {
    marginWidth = 20; // to avoid scrollbars
    marginHeight = 40;
    var selection = d3.select(this);
    if (datum.tag == "svg") {
        var width = window.innerWidth;
        var height = window.innerHeight;
        datum.attributes.width = width - marginWidth;
        datum.attributes.height = height - marginHeight;
    }
}


%%DOT_TEMPLATE%%

updateFrames();
initSlider();

</script>
"""

EXAMPLES = """

usage examples:

# Read events from 'test_trace.json' and generate 'test.html'
animate_lock_graph.py -i test_trace.json -o test.html
"""

parser = argparse.ArgumentParser(
    description="",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=EXAMPLES,
)
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"{parser.prog} ({__version__})",
)
parser.add_argument("-v", "--verbose", action="store_true", help="be verbose")
parser.add_argument(
    "-o",
    "--output",
    type=str,
    dest="output_file",
    default=None,
    help="write the trace into output file",
    required=True,
)
parser.add_argument(
    "-i",
    "--input",
    type=str,
    dest="input_file",
    default=None,
    help="the input file with the events",
    required=True,
)
parser.add_argument(
    "-f",
    "--force",
    action="store_true",
    help="overwrite the output file if it already exists",
)


class DOTModel:
    def __init__(self, input_file, verbose) -> None:
        self.input_file = input_file
        self.verbose = verbose
        self.dot_graphs = []
        self.graph = igraph.Graph(directed=True)
        self.calculate_graphs()

    def calculate_graphs(self):
        """
        Calculate the dot graphs for each line of the input
        """
        with open(self.input_file, "r", encoding="utf-8") as input_file_handle:
            for line in input_file_handle:
                json_data = json.loads(line)
                self.handle_json(json_data)

    def handle_json(self, event):
        """
        Handle the JSON encoded graph event
        """
        vertex_name_query = f"query_{event['pid']}"

        if event["event"] == "QUERY_BEGIN":
            self.graph.add_vertex(vertex_name_query, type="query", label=event["query"])
            self.generate_graph()
            return

        if event["event"] == "LOCK_GRANTED_LOCAL":
            if self.verbose:
                print(f"Processing {event}")

            tablename = StringHelper.get_tablename(event)
            lock_type = event["lock_type"]

            if not self.graph.vs.select(label_eq=tablename):
                self.graph.add_vertex(tablename, label=tablename)

            # Get numeric value of lock
            lock_numeric_value = PostgreSQLLockHelper.lock_type_to_int(lock_type)

            # Get existing edge to this table
            edge_id = self.graph.get_eid(vertex_name_query, tablename, error=False)

            # If not exist, create
            if edge_id == -1:
                lock_value = PostgreSQLLockHelper.encode_locks_into_value(
                    [lock_numeric_value]
                )
                self.graph.add_edge(vertex_name_query, tablename, lock_value=lock_value)
                self.generate_graph()
                return

            # Else, modify edge and add lock
            edge = self.graph.es[edge_id]
            decoded_lock_value = PostgreSQLLockHelper.decode_locks_from_value(
                edge["lock_value"]
            )
            decoded_lock_value.append(lock_numeric_value)
            edge["lock_value"] = PostgreSQLLockHelper.encode_locks_into_value(
                decoded_lock_value
            )

            return

        if event["event"] == "LOCK_UNGRANTED_LOCAL":
            if self.verbose:
                print(f"Processing {event}")

            tablename = StringHelper.get_tablename(event)
            lock_type = event["lock_type"]
            lock_numeric_value = PostgreSQLLockHelper.lock_type_to_int(lock_type)

            edge_id = self.graph.get_eid(vertex_name_query, tablename)
            edge = self.graph.es[edge_id]

            decoded_lock_value = PostgreSQLLockHelper.decode_locks_from_value(
                edge["lock_value"]
            )

            if lock_numeric_value in decoded_lock_value:
                decoded_lock_value.remove(lock_numeric_value)
            else:
                print(
                    f"Lock {lock_numeric_value} for table {tablename} removed but not requested"
                )

            edge["lock_value"] = PostgreSQLLockHelper.encode_locks_into_value(
                decoded_lock_value
            )

            # Delete the edge if not locks hold
            if len(decoded_lock_value) == 0:
                self.graph.delete_edges(edge)

            # Is the vertex unconnected (degree == 0) after we deleted the edge?
            vertex = self.graph.vs.select(label_eq=tablename)
            if self.graph.degree(vertex)[0] == 0:
                # self.graph.delete_vertices(tablename)
                self.graph.delete_vertices(vertex)

            self.generate_graph()
            return

    def generate_graph(self):
        """
        Generate a DOT graph based on the igraph
        """
        vertices = self.graph.vs
        edges = self.graph.es

        # Reduce space if more nodes are present
        if len(vertices) > 15:
            mindist = "0.2"
        elif len(vertices) > 10:
            mindist = "0.4"
        elif len(vertices) > 6:
            mindist = "0.75"
        elif len(vertices) > 4:
            mindist = "1.0"
        else:
            mindist = "1.7"

        dot = graphviz.Digraph("lock-graph", graph_attr={"mindist": mindist})

        for vertex in vertices:
            if vertex["type"] == "query":
                dot.node(
                    vertex["name"], vertex["label"], fillcolor="gray", style="filled"
                )
            else:
                display_name = vertex["label"]
                display_name = StringHelper.split_string(display_name, 20)
                dot.node(
                    vertex["name"],
                    display_name,
                    shape="box",
                    fillcolor="lightgray",
                    style="filled",
                )

        for edge in edges:
            # Decode graph lock value into list of lock values
            decoded_lock_value = PostgreSQLLockHelper.decode_locks_from_value(
                edge["lock_value"]
            )

            # Convert numeric lock values into a string values and join them into a single string
            label = ",".join(
                map(PostgreSQLLockHelper.lock_type_to_str, decoded_lock_value)
            )

            # Add edge to dot graph
            dot.edge(
                vertices[edge.source]["name"],
                vertices[edge.target]["name"],
                label=label,
            )

        # Convert Dot graph into string.
        # Every line needs to be quoted and line breaks need a ",""
        #
        # For example:
        #
        # ['digraph "lock-graph" {',
        # '       query_171379 [label="select count(*) from sensor_data ;"]',
        # '       "public.sensor_data" [label="public.sensor_data"]',
        # '       query_171379 -> "public.sensor_data"',
        # '}'],
        lines = dot.source.split("\n")
        lines = [f"'{line}'" for line in lines if line]
        lines = ",\n".join(lines)

        # Add dot string to list of graphs
        self.dot_graphs.append(f"[{lines}]")

    def get_html(self):
        """
        Convert all dot graphs into a single string
        """
        result = "var dots = [\n"
        graphs = ",\n".join(self.dot_graphs)
        result += graphs
        result += "];\n"
        return result


# pylint: disable=too-few-public-methods
class StringHelper:
    @staticmethod
    def split_string(mystring, max_length):
        """
        Add linebreaks to the string after max_length chars at
        the next dot or underscore.
        """

        result = ""
        chars_without_break = 0
        for element in mystring:
            result += element
            chars_without_break += 1

            if chars_without_break > max_length and element in (".", "_"):
                chars_without_break = 0
                result += "\\n"

        return result

    @staticmethod
    def get_tablename(event):
        """
        Get the tablename or the Oid of the event.
        """
        if "table" in event:
            return event["table"]

        return f"Oid {event['oid']}"


def main():
    """
    Entry point for the animate_lock_graph script
    """
    args = parser.parse_args()

    if not os.path.exists(args.input_file):
        raise ValueError(f"Input file does not exist {args.input_file}")

    if os.path.exists(args.output_file) and not args.force:
        raise ValueError(f"Output file already exists {args.output_file}")

    # Create a new dot model, process the events in the input file
    # and generate the HTML output
    dot_model_instance = DOTModel(args.input_file, args.verbose)

    with open(args.output_file, "w", encoding="utf-8") as output:
        dots = dot_model_instance.get_html()
        output_html = HTML_TEMPLATE.replace("%%DOT_TEMPLATE%%", dots)
        output.write(output_html)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# PostgreSQL lock tracer
#
# This tool traces the lock operations that PostgreSQL performs.
#
# Unfortunately, PostgreSQL does not provide USDT probes for
# these locks. Therefore BPF, UProbes, and parameter parsing
# is used to trace these events.
###############################################

import os
import sys
import json
import argparse

from abc import ABC
from enum import IntEnum, auto, unique
from bcc import BPF
from prettytable import PrettyTable

from pg_lock_tracer import __version__
from pg_lock_tracer.oid_resolver import OIDResolver
from pg_lock_tracer.helper import PostgreSQLLockHelper, BPFHelper

EXAMPLES = """

usage examples:
# Trace use binary '/home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres' for tracing and trace pid 1234
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234

# Trace two PIDs
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -p 5678

# Be verbose
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -v 

# Use the given db connection to access the catalog of PID 1234 to resolve OIDs
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -r 1234:psql://jan@localhost/test2

# Output in JSON format
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -j

# Print stacktrace on deadlock
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -s DEADLOCK

# Print stacktrace for locks and deadlocks
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -s LOCK DEADLOCK

# Trace only Transaction and Query related events
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -t TRANSACTION QUERY

# Write the output into file 'trace'
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 -o trace

# Show statistics about locks
pg_lock_tracer -x /home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres -p 1234 --statistics
"""


class TraceEvents(IntEnum):
    """
    Events to trace
    """

    TRANSACTION = auto()
    QUERY = auto()
    TABLE = auto()
    LOCK = auto()
    INVALIDATION = auto()
    ERROR = auto()


parser = argparse.ArgumentParser(
    description="",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=EXAMPLES,
)
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"{parser.prog} ({__version__})",
)
parser.add_argument("-v", "--verbose", action="store_true", help="be verbose")
parser.add_argument(
    "-j", "--json", action="store_true", help="generate output as JSON data"
)
parser.add_argument(
    "-p",
    "--pid",
    type=int,
    nargs="+",
    action="extend",
    dest="pids",
    metavar="PID",
    help="the pid(s) to trace",
)
parser.add_argument(
    "-x",
    "--exe",
    type=str,
    required=True,
    dest="path",
    metavar="PATH",
    help="path to binary",
)
parser.add_argument(
    "-r",
    "--oid-resolver",
    type=str,
    action="extend",
    default=[],
    nargs="*",
    dest="oid_resolver_urls",
    metavar="OIDResolver",
    help="OID resolver for a PID. The resolver has to be specified in format <PID:database-url>",
)
parser.add_argument(
    "-s",
    "--stacktrace",
    type=str,
    dest="stacktrace",
    action="extend",
    default=None,
    nargs="*",
    choices=["DEADLOCK", "LOCK", "UNLOCK"],
    help="print stacktrace on every of these events",
)
parser.add_argument(
    "-t",
    "--trace",
    type=str,
    dest="trace",
    action="extend",
    default=None,
    nargs="*",
    choices=[event.name for event in TraceEvents],
    help="events to trace (default: All events are traced)",
)
parser.add_argument(
    "-o",
    "--output",
    type=str,
    dest="output_file",
    default=None,
    help="write the trace into output file",
)
parser.add_argument("--statistics", action="store_true", help="print lock statistics")
parser.add_argument(
    "-d",
    "--dry-run",
    action="store_true",
    help="compile and load the BPF program but exit afterward",
)


@unique
class Events(IntEnum):
    TABLE_OPEN = 1
    TABLE_OPEN_RV = 2
    TABLE_OPEN_RV_EXTENDED = 3
    TABLE_CLOSE = 4
    ERROR = 5
    QUERY_BEGIN = 20
    QUERY_END = 21
    LOCK_RELATION_OID = 30
    LOCK_RELATION_OID_END = 31
    UNLOCK_RELATION_OID = 32
    LOCK_GRANTED = 33
    LOCK_GRANTED_FASTPATH = 34
    LOCK_GRANTED_LOCAL = 35
    LOCK_UNGRANTED = 36
    LOCK_UNGRANTED_FASTPATH = 37
    LOCK_UNGRANTED_LOCAL = 38
    TRANSACTION_BEGIN = 40
    TRANSACTION_COMMIT = 41
    TRANSACTION_ABORT = 42
    INVALIDATION_MESSAGES_ACCEPT = 50
    # Events over 1000 are handled regardless of any pid filter
    GLOBAL = 1000
    DEADLOCK = 1001


# From elog.h
@unique
class PGError(IntEnum):
    ERROR = 21
    FATAL = 22
    PANIC = 23


class LockStatisticsEntry:
    def __init__(self) -> None:
        # The number of requested locks
        self._lock_count = 0

        # The total time spend for lock requests
        self._lock_time_ns = 0

        # A list with the requested locks
        self._requested_locks = []

    @property
    def lock_count(self):
        return self._lock_count

    @lock_count.setter
    def lock_count(self, value):
        self._lock_count = value

    @property
    def lock_time_ns(self):
        return self._lock_time_ns

    @lock_time_ns.setter
    def lock_time_ns(self, value):
        self._lock_time_ns = value

    @property
    def requested_locks(self):
        return self._requested_locks

    @requested_locks.setter
    def requested_locks(self, lock_type):
        self._requested_locks.append(lock_type)


class PGLockTraceOutput(ABC):
    def __init__(self) -> None:
        super().__init__()
        self.statistics = {}
        self.bpf_instance = None
        self.bpf_stacks = None
        self.output_file = None
        self.oid_resolvers = None
        self.pids = None
        # Variables for lock timing
        self.last_lock_request_time = {}
        self.last_lock_relation = {}

    def set_context(
        self, bpf_instance, bpf_stacks, output_file, oid_resolvers, pids
    ) -> None:
        """
        Set the needed context variables
        """
        self.bpf_instance = bpf_instance
        self.bpf_stacks = bpf_stacks
        self.output_file = output_file
        self.oid_resolvers = oid_resolvers
        self.pids = pids

    def print_event(self, _cpu, data, _size):
        """
        Handle the output of the given event. Subclasses will implement
        the concrete logic.
        """

    def update_statistics(self, event, oid_value):
        """
        Add the lock call to the statistics and measure lock request time
        """
        if event.event_type == Events.LOCK_RELATION_OID:
            if oid_value not in self.statistics:
                self.statistics[oid_value] = LockStatisticsEntry()

            statistics_entry = self.statistics.get(oid_value)
            statistics_entry.lock_count += 1
            statistics_entry.requested_locks = event.mode

            self.last_lock_request_time[event.pid] = event.timestamp
            self.last_lock_relation[event.pid] = oid_value

        if event.event_type == Events.LOCK_RELATION_OID_END:
            lock_relation = self.last_lock_relation[event.pid]
            statistics_entry = self.statistics.get(lock_relation)
            statistics_entry.lock_time_ns += self.get_lock_wait_time(event)
            self.last_lock_relation[event.pid] = None

    def get_lock_wait_time(self, event):
        """
        Get the last lock wait time (LOCK_RELATION_OID updates
        last_lock_request_time).

        This method should be called on LOCK_RELATION_OID_END.
        """
        if event.event_type != Events.LOCK_RELATION_OID_END:
            return None

        return event.timestamp - self.last_lock_request_time[event.pid]

    def print_statistics(self):
        """
        Print lock statistics
        """
        print("\nLock statistics:\n================")

        # Oid lock statistics
        print("\nLocks per OID")
        table = PrettyTable(["Lock Name", "Requests", "Total Lock Request Time (ns)"])

        sorted_keys = sorted(
            self.statistics.keys(),
            key=lambda key: self.statistics.get(key).lock_count,
            reverse=True,
        )

        for key in sorted_keys:
            statistics = self.statistics[key]
            table.add_row([key, statistics.lock_count, statistics.lock_time_ns])

        print(table)

        # Lock type statistics
        print("\nLock types")
        table = PrettyTable(["Lock Type", "Number of requested locks"])

        # Map: Key = Lock type, Value = Number of requested locks
        requested_locks = {}

        # Gather per lock type statistics
        for statistics in self.statistics.values():
            for lock_type in statistics.requested_locks:
                locks = requested_locks.get(lock_type, 0) + 1
                requested_locks[lock_type] = locks

        # Print statistics
        for lock_type in sorted(requested_locks):
            locks = requested_locks[lock_type]
            lock_name = PostgreSQLLockHelper.lock_type_to_str(lock_type)
            table.add_row([lock_name, locks])

        print(table)

    def handle_output_line(self, line):
        """
        Handle a output line
        """
        if self.output_file:
            self.output_file.write(line + "\n")
        else:
            print(line)


class PGLockTraceOutputHuman(PGLockTraceOutput):
    # pylint: disable=too-many-branches, too-many-statements
    def print_event(self, _cpu, data, _size):
        """
        Print event in a human readable format
        """
        event = self.bpf_instance["lockevents"].event(data)

        if (
            self.pids
            and event.pid not in self.pids
            and event.event_type < Events.GLOBAL
        ):
            return

        print_prefix = f"{event.timestamp} [Pid {event.pid}]"

        # Resolve the OID to a name
        if event.object > 0:
            tablename = event.object

        # Resolve the OID to a table name
        if event.pid in self.oid_resolvers and event.object:
            resolver = self.oid_resolvers[event.pid]
            oid_value = resolver.resolve_oid(event.object)
            tablename = f"{tablename} ({oid_value})"
            self.update_statistics(event, oid_value)
        else:
            self.update_statistics(event, event.object)

        output = None
        if event.event_type == Events.TABLE_OPEN:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Table open {tablename} {lock_type}"
        elif event.event_type in (Events.TABLE_OPEN_RV, Events.TABLE_OPEN_RV_EXTENDED):
            # Table is opened using a (string) range value
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            schema = event.payload_str1.decode("utf-8")
            table = event.payload_str2.decode("utf-8")
            tablename = f"{schema}.{table}"
            output = (
                f"{print_prefix} Table open (by range value) {tablename} {lock_type}"
            )
        elif event.event_type == Events.TABLE_CLOSE:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Table close {tablename} {lock_type}"
        elif event.event_type == Events.LOCK_RELATION_OID:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Lock object {tablename} {lock_type}"
        elif event.event_type == Events.LOCK_RELATION_OID_END:
            lock_time = self.get_lock_wait_time(event)
            output = f"{print_prefix} Lock was acquired in {lock_time} ns"
        elif event.event_type == Events.UNLOCK_RELATION_OID:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Unlock relation {tablename} {lock_type}"
        elif event.event_type == Events.LOCK_GRANTED:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = (
                f"{print_prefix} Lock granted {tablename} {lock_type} "
                f"(Requested locks {event.requested})"
            )
        elif event.event_type == Events.LOCK_GRANTED_FASTPATH:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Lock granted (fastpath) {tablename} {lock_type}"
        elif event.event_type == Events.LOCK_GRANTED_LOCAL:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = (
                f"{print_prefix} Lock granted (local) {tablename} {lock_type} "
                f"(Already hold local {event.lock_local_hold})"
            )
        elif event.event_type == Events.LOCK_UNGRANTED:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = (
                f"{print_prefix} Lock ungranted {tablename} {lock_type} "
                f"(Requested locks {event.requested})"
            )
        elif event.event_type == Events.LOCK_UNGRANTED_FASTPATH:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = f"{print_prefix} Lock ungranted (fastpath) {tablename} {lock_type}"
        elif event.event_type == Events.LOCK_UNGRANTED_LOCAL:
            lock_type = PostgreSQLLockHelper.lock_type_to_str(event.mode)
            output = (
                f"{print_prefix} Lock ungranted (local) {tablename} {lock_type} "
                f"(Hold local {event.lock_local_hold})"
            )
        elif event.event_type == Events.INVALIDATION_MESSAGES_ACCEPT:
            output = f"{print_prefix} Accept invalidation messages"
        elif event.event_type == Events.ERROR:
            pgerror_value = PGError(event.mode).name
            output = f"{print_prefix} Error occurred servity: {pgerror_value}"
        elif event.event_type == Events.QUERY_BEGIN:
            query = event.payload_str1.decode("utf-8")
            output = f"{print_prefix} Query begin '{query}'"
        elif event.event_type == Events.QUERY_END:
            output = f"{print_prefix} Query done\n"
        elif event.event_type == Events.TRANSACTION_BEGIN:
            output = f"{print_prefix} Transaction begin"
        elif event.event_type == Events.TRANSACTION_COMMIT:
            output = f"{print_prefix} Transaction commit"
        elif event.event_type == Events.TRANSACTION_ABORT:
            output = f"{print_prefix} Transaction abort"
        elif event.event_type == Events.DEADLOCK:
            output = f"{print_prefix} DEADLOCK DETECTED"
        else:
            raise ValueError(f"Unsupported event type {event.event_type}")

        self.handle_output_line(output)
        self.print_stacktace_if_available(event)

    def print_stacktace_if_available(self, event):
        """
        Print the stacktrace of an event if available
        """
        if event.stackid == 0 or self.bpf_stacks is None:
            return

        if event.stackid < 0:
            print(
                "Error stack is missing. Try to increase BPF_STACK_TRACE buffer size."
            )
        else:
            for frame in self.bpf_stacks.walk(event.stackid):
                line = self.bpf_instance.sym(
                    frame, event.pid, show_offset=True, show_module=True
                )
                line = line.decode("utf-8")
                # Get line with: 'gdb info line *(symbol+0x1111)'
                line = f"\t{line}"
                self.handle_output_line(line)


class PGLockTraceOutputJSON(PGLockTraceOutput):
    def print_event(self, _cpu, data, _size):
        """
        Print event in JSON format
        """
        event = self.bpf_instance["lockevents"].event(data)

        if (
            self.pids
            and event.pid not in self.pids
            and event.event_type < Events.GLOBAL
        ):
            return

        output = {}
        output["timestamp"] = event.timestamp
        output["pid"] = event.pid
        output["event"] = Events(event.event_type).name

        if event.event_type in (
            Events.TABLE_OPEN,
            Events.TABLE_OPEN_RV,
            Events.TABLE_OPEN_RV_EXTENDED,
            Events.TABLE_CLOSE,
            Events.LOCK_RELATION_OID,
            Events.UNLOCK_RELATION_OID,
            Events.LOCK_GRANTED,
            Events.LOCK_GRANTED_FASTPATH,
            Events.LOCK_GRANTED_LOCAL,
            Events.LOCK_UNGRANTED,
            Events.LOCK_UNGRANTED_FASTPATH,
            Events.LOCK_UNGRANTED_LOCAL,
        ):
            output["lock_type"] = PostgreSQLLockHelper.lock_type_to_str(event.mode)

        # Resolve OID to tablename
        if event.pid in self.oid_resolvers and event.object:
            resolver = self.oid_resolvers[event.pid]
            oid_value = resolver.resolve_oid(event.object)
            output["table"] = oid_value
            self.update_statistics(event, oid_value)
        else:
            self.update_statistics(event, event.object)

        # Resolve the OID to a name
        if event.object:
            output["oid"] = event.object

        if event.event_type == Events.ERROR:
            pgerror_value = PGError(event.mode).name
            output["servity"] = pgerror_value
        elif event.event_type == Events.QUERY_BEGIN:
            output["query"] = event.payload_str1.decode("utf-8")
        elif event.event_type in (Events.TABLE_OPEN_RV, Events.TABLE_OPEN_RV_EXTENDED):
            # Table is opened using a (string) range value
            schema = event.payload_str1.decode("utf-8")
            table = event.payload_str2.decode("utf-8")
            output["table"] = f"{schema}.{table}"
        elif event.event_type == Events.LOCK_GRANTED_LOCAL:
            output["lock_local_hold"] = event.lock_local_hold
        elif event.event_type == Events.LOCK_RELATION_OID_END:
            lock_time = self.get_lock_wait_time(event)
            output["lock_time"] = lock_time

        self.add_stacktrace_if_available(output, event)

        self.handle_output_line(json.dumps(output))

    def add_stacktrace_if_available(self, output, event):
        """
        Add a stacktrace to the JSON structure if available
        """
        if event.stackid == 0 or self.bpf_stacks is None:
            return

        if event.stackid < 0:
            output["stacktrace"] = "MISSING"
        else:
            lines = []

            # Get stacktrace symbol with module
            for frame in self.bpf_stacks.walk(event.stackid):
                line = self.bpf_instance.sym(
                    frame, event.pid, show_offset=True, show_module=True
                )
                lines.append(line.decode("utf-8"))

            # Merge lines into a single string
            stacktrace = ", ".join(lines)
            output["stacktrace"] = stacktrace


class PGLockTracer:
    def __init__(self, prog_args):
        self.bpf_instance = None
        self.bpf_stacks = None
        self.output_file = None
        self.output_class = None
        self.args = prog_args

        # A map of OID resolvers. One resolver per PID is needed
        # because the Oid depend on the catalog of the database.
        self.oid_resolvers = {}

        for oid_resolver_url in self.args.oid_resolver_urls:
            if ":" not in oid_resolver_url:
                raise ValueError(
                    f"Resolver URL has to be in format: 'PID:URL' ({oid_resolver_url} was provided)"
                )

            split_url = oid_resolver_url.split(":", 1)
            resolver_pid = int(split_url[0])
            database_url = split_url[1]

            if resolver_pid not in self.args.pids:
                print(
                    f"Specified resolver for PID {resolver_pid}, but PID is not monitored"
                )
                sys.exit(1)

            if self.args.verbose:
                print(f"Add resolver for PID {resolver_pid} with URL {database_url}")

            oid_resolver = OIDResolver(database_url)
            self.oid_resolvers[resolver_pid] = oid_resolver

        # Belong the processes to the binary?
        BPFHelper.check_pid_exe(self.args.pids, self.args.path)

        # Does the output file already exists?
        if self.args.output_file and os.path.exists(self.args.output_file):
            raise ValueError(f"Output file {self.args.output_file} already exists")

    @staticmethod
    def generate_c_defines(stacktrace_events, verbose):
        """
        Create C defines from python enums
        """
        enum_defines = BPFHelper.enum_to_defines(Events, "EVENT")
        error_defines = BPFHelper.enum_to_defines(PGError, "PGERROR")
        defines = enum_defines + error_defines

        # Print stacktrace for each lock
        if stacktrace_events and "LOCK" in stacktrace_events:
            defines += "#define STACKTRACE_LOCK\n"
            if verbose:
                print("Print stacktrace on each lock event")

        # Print stacktrace for each unlock
        if stacktrace_events and "UNLOCK" in stacktrace_events:
            defines += "#define STACKTRACE_UNLOCK\n"
            if verbose:
                print("Print stacktrace on each unlock event")

        # Print stacktrace on deadlock
        if stacktrace_events and "DEADLOCK" in stacktrace_events:
            defines += "#define STACKTRACE_DEADLOCK\n"
            if verbose:
                print("Print stacktrace on each deadlock")

        return defines

    def init(self):
        """
        Init the PostgreSQL lock tracer
        """

        defines = PGLockTracer.generate_c_defines(
            self.args.stacktrace, self.args.verbose
        )
        bpf_program = BPFHelper.read_bpf_program("pg_lock_tracer.c")
        bpf_program_final = bpf_program.replace("__DEFINES__", defines)

        if self.args.verbose:
            print(bpf_program_final)

        # Disable warnings like
        # 'warning: '__HAVE_BUILTIN_BSWAP32__' macro redefined [-Wmacro-redefined]'
        bpf_cflags = ["-Wno-macro-redefined"] if not self.args.verbose else []

        print("===> Compiling BPF program")
        self.bpf_instance = BPF(text=bpf_program_final, cflags=bpf_cflags)

        print("===> Attaching BPF probes")
        self.attach_probes()

        # Stack traces requested?
        if self.args.stacktrace:
            self.bpf_stacks = self.bpf_instance.get_table("stacks")

        # Open file for output if provided
        if self.args.output_file:
            # pylint: disable=consider-using-with
            self.output_file = open(self.args.output_file, "w", encoding="utf-8")
            if not self.output_file.writable():
                raise ValueError(
                    f"Output file {self.args.output_file} is not writeable"
                )

        # Output as human readable text or as json?
        self.output_class = (
            PGLockTraceOutputJSON() if self.args.json else PGLockTraceOutputHuman()
        )

        # Init the output class
        self.output_class.set_context(
            self.bpf_instance,
            self.bpf_stacks,
            self.output_file,
            self.oid_resolvers,
            self.args.pids,
        )

        # Open the event queue
        self.bpf_instance["lockevents"].open_perf_buffer(
            self.output_class.print_event, page_cnt=BPFHelper.page_cnt
        )

    def attach_probes(self):
        """
        Attach BPF probes
        """
        # Transaction probes
        if self.args.trace is None or TraceEvents.TRANSACTION.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^StartTransaction$",
                "bpf_transaction_begin",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^CommitTransaction$",
                "bpf_transaction_commit",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^AbortTransaction$",
                "bpf_transaction_abort",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^DeadLockReport$",
                "bpf_deadlock",
                self.args.verbose,
            )

        # Query probes
        if self.args.trace is None or TraceEvents.QUERY.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^exec_simple_query$",
                "bpf_query_begin",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^exec_simple_query$",
                "bpf_query_end",
                self.args.verbose,
                False,
            )

        # Table probes
        if self.args.trace is None or TraceEvents.TABLE.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^table_open$",
                "bpf_table_open",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^table_openrv$",
                "bpf_table_openrv",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^table_openrv_extended$",
                "bpf_table_openrv_extended",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^table_close$",
                "bpf_table_close",
                self.args.verbose,
            )

        # Lock probes
        if self.args.trace is None or TraceEvents.LOCK.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^LockRelationOid$",
                "bpf_lock_relation_oid",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^LockRelationOid$",
                "bpf_lock_relation_oid_end",
                self.args.verbose,
                False,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^UnlockRelationOid$",
                "bpf_unlock_relation_oid",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^GrantLock$",
                "bpf_lock_grant",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^FastPathGrantRelationLock$",
                "bpf_lock_fastpath_grant",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^GrantLockLocal$",
                "bpf_lock_local_grant",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^UnGrantLock$",
                "bpf_lock_ungrant",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^FastPathUnGrantRelationLock$",
                "bpf_lock_fastpath_ungrant",
                self.args.verbose,
            )
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^RemoveLocalLock$",
                "bfp_local_lock_ungrant",
                self.args.verbose,
            )

        # Invalidation messages probes
        if self.args.trace is None or TraceEvents.INVALIDATION.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^AcceptInvalidationMessages$",
                "bpf_accept_invalidation_messages",
                self.args.verbose,
            )

        # Error probes
        if self.args.trace is None or TraceEvents.ERROR.name in self.args.trace:
            BPFHelper.register_ebpf_probe(
                self.args.path,
                self.bpf_instance,
                "^errstart$",
                "bpf_errstart",
                self.args.verbose,
            )

    def run(self):
        """
        Run the BPF program and read results
        """

        print("===> Ready to trace queries")
        while True:
            try:
                self.bpf_instance.perf_buffer_poll()
            except KeyboardInterrupt:
                if self.output_file:
                    self.output_file.close()

                if self.args.statistics:
                    self.output_class.print_statistics()
                sys.exit(0)


def main():
    """
    Entry point for the BPF based PostgreSQL lock tracer.
    """
    args = parser.parse_args()

    pg_lock_tracer = PGLockTracer(args)
    pg_lock_tracer.init()

    if not args.dry_run:
        pg_lock_tracer.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# PostgreSQL spinlock delay tracer.
#
###############################################

import sys
import argparse

from bcc import BPF

from pg_lock_tracer import __version__
from pg_lock_tracer.helper import BPFHelper

EXAMPLES = """examples:
# Trace spin delays of the given PostgreSQL binary
pg_spinlock_delay_tracer -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace spin delays of the PID 1234
pg_spinlock_delay_tracer -p 1234 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace spin delays of the PID 1234 and 5678
pg_spinlock_delay_tracer -p 1234 -p 5678 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres

# Trace spin delays of the PID 1234 and be verbose
pg_spinlock_delay_tracer -p 1234 -x /home/jan/postgresql-sandbox/bin/REL_14_9_DEBUG/bin/postgres -v
"""

parser = argparse.ArgumentParser(
    description="",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=EXAMPLES,
)
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"{parser.prog} ({__version__})",
)
parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
parser.add_argument(
    "-p",
    "--pid",
    type=int,
    nargs="+",
    dest="pids",
    metavar="PID",
    help="the pid(s) to trace",
)
parser.add_argument(
    "-x",
    "--exe",
    type=str,
    required=True,
    dest="path",
    metavar="PATH",
    help="path to binary",
)
parser.add_argument(
    "-d",
    "--dry-run",
    action="store_true",
    help="compile and load the BPF program but exit afterward",
)


class PGSpinDelayTracer:
    def __init__(self, prog_args):
        self.bpf_instance = None
        self.args = prog_args

        # Belong the processes to the binary?
        BPFHelper.check_pid_exe(self.args.pids, self.args.path)

    @staticmethod
    def _decode_field(value):
        return value.decode("utf-8", "replace").rstrip("\x00")

    def print_lock_event(self, _cpu, data, _size):
        """
        Print a new spin delay event.
        """
        event = self.bpf_instance["lockevents"].event(data)

        if self.args.pids and event.pid not in self.args.pids:
            return

        file_name = self._decode_field(event.file) or "(unknown)"
        func_name = self._decode_field(event.func) or "(unknown)"

        print(
            f"{event.timestamp} [Pid {event.pid}] SpinDelay "
            f"spins={event.spins} delays={event.delays} "
            f"cur_delay={event.cur_delay} at {func_name}, {file_name}:{event.line}"
        )

    def init(self):
        """
        Init the PostgreSQL spin delay tracer
        """
        bpf_program = BPFHelper.read_bpf_program("pg_spinlock_delay_tracer.c")

        if self.args.verbose:
            print(bpf_program)

        # Disable warnings like
        # 'warning: '__HAVE_BUILTIN_BSWAP32__' macro redefined [-Wmacro-redefined]'
        bpf_cflags = ["-Wno-macro-redefined"] if not self.args.verbose else []

        print("===> Compiling BPF program")
        self.bpf_instance = BPF(text=bpf_program, cflags=bpf_cflags)

        print("===> Attaching BPF probes")
        self.attach_probes()

        # Open the event queue
        self.bpf_instance["lockevents"].open_perf_buffer(
            self.print_lock_event, page_cnt=BPFHelper.page_cnt
        )

    def attach_probes(self):
        """
        Attach BPF probes
        """
        BPFHelper.register_ebpf_probe(
            self.args.path,
            self.bpf_instance,
            "^perform_spin_delay$",
            "spin_delay",
            self.args.verbose,
        )

    def run(self):
        """
        Run the BPF program and read results
        """
        print("===> Ready to trace")
        while True:
            try:
                self.bpf_instance.perf_buffer_poll()
            except KeyboardInterrupt:
                sys.exit(0)


def main():
    """
    Entry point for the BPF based PostgreSQL spin delay tracer.
    """
    args = parser.parse_args()

    pg_spin_delay_tracer = PGSpinDelayTracer(args)
    pg_spin_delay_tracer.init()

    if not args.dry_run:
        pg_spin_delay_tracer.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# PostgreSQL LW lock tracer. To use this script, PostgreSQL has to be
# compiled with '--enable-dtrace'.
#
# See https://www.postgresql.org/docs/current/dynamic-trace.html
#
# List all available USDT probes
# sudo bpftrace -l "usdt:/home/jan/postgresql-sandbox/bin/REL_15_1_DEBUG/bin/postgres:*"
###############################################

import sys
import argparse

from enum import IntEnum, unique

from bcc import BPF, USDT
from prettytable import PrettyTable

from pg_lock_tracer import __version__
from pg_lock_tracer.helper import BPFHelper

EXAMPLES = """examples:
# Trace the LW locks of the PID 1234
pg_lw_lock_tracer -p 1234

# Trace the LW locks of the PIDs 1234 and 5678
pg_lw_lock_tracer -p 1234 -p 5678

# Trace the LW locks of the PID 1234 and be verbose
pg_lw_lock_tracer -p 1234 -v

# Trace the LW locks of the PID 1234 and collect statistics
pg_lw_lock_tracer -p 1234 -v --statistics
"""

parser = argparse.ArgumentParser(
    description="",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    epilog=EXAMPLES,
)
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"{parser.prog} ({__version__})",
)
parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
parser.add_argument(
    "-p",
    "--pid",
    type=int,
    nargs="+",
    dest="pids",
    metavar="PID",
    help="the pid(s) to trace",
    required=True,
)
parser.add_argument(
    "-d",
    "--dry-run",
    action="store_true",
    help="compile and load the BPF program but exit afterward",
)
parser.add_argument("--statistics", action="store_true", help="print lock statistics")


@unique
class Events(IntEnum):
    LOCK = 0
    LOCK_OR_WAIT = 1
    LOCK_OR_WAIT_FAIL = 2
    UNLOCK = 3
    WAIT_START = 4
    WAIT_DONE = 5
    COND_ACQUIRE = 6
    COND_ACQUIRE_FAIL = 7


@unique
class LWLockMode(IntEnum):
    LW_EXCLUSIVE = 0
    LW_SHARED = 1
    LW_WAIT_UNTIL_FREE = 2


class LockStatisticsEntry:
    def __init__(self) -> None:
        # The number of non-waited requested locks
        self._direct_lock_count = 0

        # The number of acquire lock or wait calls
        self._acquire_or_wait_count = 0

        # The number of failed acquire lock or wait calls
        self._acquire_or_wait_failed_count = 0

        # The number of locks with condition
        self._lock_cond_count = 0

        # The number of failed lock with condition
        self._lock_cond_failed_count = 0

        # The number of lock waits
        self._wait_lock_count = 0

        # The total time spend for lock wait requests
        self._lock_wait_time_ns = 0

        # A list with the requested locks
        self._requested_locks = []

    @property
    def direct_lock_count(self):
        return self._direct_lock_count

    @direct_lock_count.setter
    def direct_lock_count(self, value):
        self._direct_lock_count = value

    @property
    def acquire_or_wait_count(self):
        return self._acquire_or_wait_count

    @acquire_or_wait_count.setter
    def acquire_or_wait_count(self, value):
        self._acquire_or_wait_count = value

    @property
    def acquire_or_wait_failed_count(self):
        return self._acquire_or_wait_failed_count

    @acquire_or_wait_failed_count.setter
    def acquire_or_wait_failed_count(self, value):
        self._acquire_or_wait_failed_count = value

    @property
    def wait_lock_count(self):
        return self._wait_lock_count

    @wait_lock_count.setter
    def wait_lock_count(self, value):
        self._wait_lock_count = value

    @property
    def lock_cond_count(self):
        return self._lock_cond_count

    @lock_cond_count.setter
    def lock_cond_count(self, value):
        self._lock_cond_count = value

    @property
    def lock_cond_failed_count(self):
        return self._lock_cond_failed_count

    @lock_cond_failed_count.setter
    def lock_cond_failed_count(self, value):
        self._lock_cond_failed_count = value

    @property
    def lock_wait_time_ns(self):
        return self._lock_wait_time_ns

    @lock_wait_time_ns.setter
    def lock_wait_time_ns(self, value):
        self._lock_wait_time_ns = value

    @property
    def requested_locks(self):
        return self._requested_locks

    @requested_locks.setter
    def requested_locks(self, lock_type):
        self._requested_locks.append(lock_type)


class PGLWLockTracer:
    def __init__(self, prog_args):
        self.bpf_instance = None
        self.usdts = None
        self.prog_args = prog_args
        self.statistics = {}

        # Variables for lock timing
        self.last_lock_request_time = {}

    def update_statistics(self, event, tranche, lock_mode):
        """
        Update the statistics
        """

        if tranche not in self.statistics:
            self.statistics[tranche] = LockStatisticsEntry()

        statistics_entry = self.statistics.get(tranche)

        # Lock directly requested
        if event.event_type == Events.LOCK:
            statistics_entry.direct_lock_count += 1
            statistics_entry.requested_locks = lock_mode
            return

        # LWLockAcquireOrWait - Acquired
        if event.event_type == Events.LOCK_OR_WAIT:
            statistics_entry.acquire_or_wait_count += 1
            statistics_entry.requested_locks = lock_mode
            return

        # LWLockAcquireOrWait - Waited
        if event.event_type == Events.LOCK_OR_WAIT_FAIL:
            statistics_entry.acquire_or_wait_failed_count += 1
            statistics_entry.requested_locks = lock_mode
            return

        # Wait for lock
        if event.event_type == Events.WAIT_START:
            statistics_entry.wait_lock_count += 1
            self.last_lock_request_time[event.pid] = event.timestamp
            return

        # Wait for lock done
        if event.event_type == Events.WAIT_DONE:
            wait_time = self.get_lock_wait_time(event)
            statistics_entry.lock_wait_time_ns += wait_time
            return

        # LWLockConditionalAcquire - Acquire with condition
        if event.event_type == Events.COND_ACQUIRE:
            statistics_entry.lock_cond_count += 1
            statistics_entry.requested_locks = lock_mode
            return

        # LWLockConditionalAcquire - Condition not possible
        if event.event_type == Events.COND_ACQUIRE_FAIL:
            statistics_entry.lock_cond_failed_count += 1
            statistics_entry.requested_locks = lock_mode
            return

    def get_lock_wait_time(self, event):
        """
        Get the last lock wait time (WAIT_START updates
        last_lock_request_time).
        """
        if event.event_type != Events.WAIT_DONE:
            return None

        return event.timestamp - self.last_lock_request_time[event.pid]

    def print_lock_event(self, _cpu, data, _size):
        """
        Print a new lock event.

        Developer note:
        Wait events can be tested with second PostgreSQL process and gdb
        b LWLockAcquireOrWait
        """
        event = self.bpf_instance["lockevents"].event(data)
        tranche = event.tranche.decode("utf-8")

        print_prefix = f"{event.timestamp} [Pid {event.pid}]"
        lock_mode = LWLockMode(event.mode).name

        self.update_statistics(event, tranche, lock_mode)

        if event.event_type == Events.LOCK:
            print(
                f"{print_prefix} Acquired lock {tranche} (mode {lock_mode}) / LWLockAcquire()"
            )
        elif event.event_type == Events.LOCK_OR_WAIT:
            print(
                f"{print_prefix} Acquired lock {tranche} (mode {lock_mode}) "
                "/ LWLockConditionalAcquire()"
            )
        elif event.event_type == Events.LOCK_OR_WAIT_FAIL:
            print(
                f"{print_prefix} Waited but not acquired {tranche} (mode {lock_mode}) "
                "/ LWLockConditionalAcquire()"
            )
        elif event.event_type == Events.UNLOCK:
            print(f"{print_prefix} Unlock {tranche}")
        elif event.event_type == Events.WAIT_START:
            print(f"{print_prefix} Wait for {tranche}")
        elif event.event_type == Events.WAIT_DONE:
            lock_time = self.get_lock_wait_time(event)
            print(f"{print_prefix} Wait for {tranche} lock took {lock_time} ns")
        elif event.event_type == Events.COND_ACQUIRE:
            print(
                f"{print_prefix} Acquired lock {tranche} (mode {lock_mode}) "
                "/ LWLockConditionalAcquire()"
            )
        elif event.event_type == Events.COND_ACQUIRE_FAIL:
            print(
                f"{print_prefix} Failed to acquire lock {tranche} (mode {lock_mode}) "
                "/ LWLockConditionalAcquire()"
            )
        else:
            raise ValueError(f"Unknown event type {event.event_type}")

    def init(self):
        """
        Compile and load the BPF program
        """
        print(f"==> Attaching to PIDs {self.prog_args.pids}")
        self.usdts = list(map(lambda pid: USDT(pid=pid), self.prog_args.pids))

        # See https://www.postgresql.org/docs/15/dynamic-trace.html
        for usdt in self.usdts:
            usdt.enable_probe("lwlock__acquire", "lwlock_acquire")
            usdt.enable_probe("lwlock__acquire__or__wait", "lwlock_acquire_or_wait")
            usdt.enable_probe(
                "lwlock__acquire__or__wait__fail", "lwlock_acquire_or_wait_fail"
            )
            usdt.enable_probe("lwlock__release", "lwlock_release")
            usdt.enable_probe("lwlock__wait__start", "lwlock_wait_start")
            usdt.enable_probe("lwlock__wait__done", "lwlock_wait_done")
            usdt.enable_probe("lwlock__condacquire", "lwlock_condacquire")
            usdt.enable_probe("lwlock__condacquire__fail", "lwlock_condacquire_fail")

        if self.prog_args.verbose:
            print("=======")
            print("\n".join(map(lambda u: u.get_text(), self.usdts)))
            print("=======")

        enum_defines = BPFHelper.enum_to_defines(Events, "EVENT")

        bpf_program = BPFHelper.read_bpf_program("pg_lw_lock_tracer.c")
        bpf_program_final = bpf_program.replace("__DEFINES__", enum_defines)

        if self.prog_args.verbose:
            print(bpf_program_final)

        # Disable warnings like
        # 'warning: '__HAVE_BUILTIN_BSWAP32__' macro redefined [-Wmacro-redefined]'
        bpf_cflags = ["-Wno-macro-redefined"] if not self.prog_args.verbose else []

        print("===> Compiling BPF program")
        self.bpf_instance = BPF(
            text=bpf_program_final, cflags=bpf_cflags, usdt_contexts=self.usdts
        )

        self.bpf_instance["lockevents"].open_perf_buffer(
            self.print_lock_event, page_cnt=BPFHelper.page_cnt
        )

    def print_statistics(self):
        """
        Print lock statistics
        """
        print("\nLock statistics:\n================")

        # Tranche lock statistics
        print("\nLocks per tranche")
        table = PrettyTable(
            [
                "Tranche",
                "Acquired",
                "AcquireOrWait (Acquired)",
                "AcquireOrWait (Waited)",
                "ConditionalAcquire (Acquired)",
                "ConditionalAcquire (Failed)",
                "Waits",
                "Wait time (ns)",
            ]
        )

        for key in sorted(self.statistics):
            statistics = self.statistics[key]
            table.add_row(
                [
                    key,
                    statistics.direct_lock_count,
                    statistics.acquire_or_wait_count,
                    statistics.acquire_or_wait_failed_count,
                    statistics.lock_cond_count,
                    statistics.lock_cond_failed_count,
                    statistics.wait_lock_count,
                    statistics.lock_wait_time_ns,
                ]
            )

        print(table)

        # Type lock statistics
        print("\nLocks per type")
        table = PrettyTable(["Lock type", "Requests"])

        # Map: Key = Lock type, Value = Number of requested locks
        requested_locks = {}

        for statistics in self.statistics.values():
            for lock_type in statistics.requested_locks:
                locks = requested_locks.get(lock_type, 0) + 1
                requested_locks[lock_type] = locks

        for lock_type in sorted(requested_locks):
            locks = requested_locks[lock_type]
            table.add_row([lock_type, locks])

        print(table)

    def run(self):
        """
        Run the BPF program and read results
        """
        print("===> Ready to trace")
        while True:
            try:
                self.bpf_instance.perf_buffer_poll()
            except KeyboardInterrupt:
                if self.prog_args.statistics:
                    self.print_statistics()
                sys.exit(0)


def main():
    """
    Entry point for the BPF based PostgreSQL LW lock tracer.
    """
    args = parser.parse_args()

    pg_lock_tracer = PGLWLockTracer(args)
    pg_lock_tracer.init()

    if not args.dry_run:
        pg_lock_tracer.run()


if __name__ == "__main__":
    main()
